from pymoode.algorithms.base.evolutionary import EvolutionaryAlgorithm
from pymoode.algorithms.base.differential import DifferentialEvolution
from pymoode.algorithms.base.genetic import GeneticAlgorithm


# Native
from typing import Optional, Tuple, Union
from abc import abstractmethod

# External
import numpy as np

# pymoo imports
from pymoo.core.population import Population
from pymoo.core.problem import Problem

# pymoode imports
from pymoode.operators.deop import DifferentialOperator



# =========================================================================================================
# Implementation
# =========================================================================================================

class DifferentialMutation(DifferentialOperator):

    def __init__(self,
                 F=None,
                 gamma=1e-4,
                 de_repair="bounce-back",
                 n_diffs=1,
                 **kwargs):
        """Differential Evolution mutation
        (Implemented similar to pymoo crossover)

        Parameters
        ----------
        F : iterable of float or float, optional
            Scale factor or mutation parameter. Defined in the range (0, 2]
            To reinforce exploration, use higher values; for exploitation, use lower values.

        gamma : float, optional
            Jitter deviation parameter. Should be in the range (0, 2). Defaults to 1e-4.

        de_repair : str, optional
            Repair of DE mutant vectors. Is either callable or one of:

                - 'bounce-back'
                - 'midway'
                - 'rand-init'
                - 'to-bounds'

            If callable, has the form fun(X, Xb, xl, xu) in which X contains mutated vectors including violations,
            Xb contains reference vectors for repair in feasible space,
            xl is a 1d vector of lower bounds, and xu a 1d vector of upper bounds.
            Defaults to 'bounce-back'.

        n_diffs : int, optional
            Number of difference vectors in operation, by default 1

        Raises
        ------
        KeyError
            Bad de_repair input
        """
        # __init__ operator
        n_parents = 1 + 2 * n_diffs
        super().__init__(n_parents=n_parents, **kwargs)

        # Other attributes
        self.F = F
        self.gamma = gamma
        self.de_repair = de_repair

    def do(
        self,
        problem: Problem,
        pop: Population,
        parents: Optional[np.ndarray] = None,
        **kwargs
    ):
        """Perform Differential Evolution mutation

        Parameters
        ----------
        problem : Problem
            Optimization problem

        pop : Population
            Original parent population at a given generation

        parents : numpy.ndarray (n_offsprings, n_select) of dtype int | None, optional
            Parent (indexes) to crossover.
            If None, the argument pop should already be sampled as parents to crossover.
            By default None

        Returns
        -------
        Population
            Offspring mutants V
        """
        # Convert pop if parents is not None
        pop, X = self.default_prepare(pop, parents)

        # Create mutation vectors
        V = self._do(problem, X, **kwargs)

        # If the problem has boundaries to be considered
        if problem.has_bounds():

            # Do de_repair
            V = self.de_repair(V, X[0], *problem.bounds())

        return Population.new("X", V)

    @abstractmethod
    def _do(
        self,
        problem: Problem,
        X: np.ndarray,
        **kwargs
    ):
        pass


class DEM(DifferentialMutation):

    def __init__(self,
                 F: Optional[Union[float, Tuple[float, float]]] = None,
                 gamma: float = 1e-4,
                 de_repair: str = "bounce-back",
                 n_diffs: int = 1,
                 **kwargs):

        # Default value for F
        if F is None:
            F = (0.0, 1.0)

        # Define which method will be used to generate F values
        if hasattr(F, "__iter__"):
            self.scale_factor = self._randomize_scale_factor
        else:
            self.scale_factor = self._scalar_scale_factor

        # Define which method will be used to generate F values
        if not hasattr(de_repair, "__call__"):
            try:
                de_repair = REPAIRS[de_repair]
            except Exception:
                raise KeyError("Repair must be either callable or in " + str(list(REPAIRS.keys())))

        # Define which strategy of rotation will be used
        if gamma is None:
            self.get_diff = self._diff_simple
        else:
            self.get_diff = self._diff_jitter

        super().__init__(
            F=F, gamma=gamma, de_repair=de_repair,
            n_diffs=n_diffs, **kwargs,
        )

    def _do(self, problem: Problem, X: np.ndarray, **kwargs):
        return self.de_mutation(X, return_differentials=False)

    def de_mutation(self, X: np.ndarray, return_differentials=True):
        """Perform DE mutation operator

        Parameters
        ----------
        X : numpy.ndarray (n_parents, n_matings, n_var)
            Three dimensional array of parents selected to take part in DE mutation

        return_differentials : bool, optional
            Either or not to return a tuple of ``(V, diffs)``, by default True

        Returns
        -------
        np.array (n_matings, n_var) | tuple
            Mutant vectors created or pair of both mutant vectors and perturbation vectors
        """

        n_parents, n_matings, n_var = X.shape
        assert n_parents % 2 == 1, "For the differential an odd number of values need to be provided"

        # Build the pairs for the differentials
        pairs = (np.arange(n_parents - 1) + 1).reshape(-1, 2)

        # The differentials from each pair subtraction
        diffs = self.get_diffs(X, pairs, n_matings, n_var)

        # Add the difference vectors to the base vector
        V = X[0] + diffs

        if return_differentials:
            return V, diffs
        else:
            return V

    def _randomize_scale_factor(self, n_matings):
        return (self.F[0] + np.random.random(n_matings) * (self.F[1] - self.F[0]))

    def _scalar_scale_factor(self, n_matings):
        return np.full(n_matings, self.F)

    def _diff_jitter(self, F, Xi, Xj, n_matings, n_var):
        F = F[:, None] * (1 + self.gamma * (np.random.random((n_matings, n_var)) - 0.5))
        return F * (Xi - Xj)

    def _diff_simple(self, F, Xi, Xj, n_matings, n_var):
        return F[:, None] * (Xi - Xj)

    def get_diffs(
        self,
        X: np.ndarray,
        pairs,
        n_matings,
        n_var
    ):

        # The differentials from each pair subtraction
        diffs = np.zeros((n_matings, n_var))

        # For each difference
        for i, j in pairs:

            # Obtain F randomized in range
            F = self.scale_factor(n_matings)

            # New difference vector
            diff = self.get_diff(F, X[i], X[j], n_matings, n_var)

            # Add the difference to the first vector
            diffs = diffs + diff

        return diffs


# =========================================================================================================
# Repairs
# =========================================================================================================


def bounce_back(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements between reference points and bounds

    Parameters
    ----------
    X : numpy.ndarray (n_samples, n_var)
        Original population in decision space (including violations)

    Xb : numpy.ndarray (n_samples, n_var)
        Feasible reference points in decision space

    xl : numpy.ndarray (n_var,)
        Lower bounds for decision variables

    xu : numpy.ndarray (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.ndarray (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - Xb[i, j])

    return X


def midway(X, Xb, xl, xu):
    """Repair strategy that sets violated elements to midpoint between reference points and bounds

    Parameters
    ----------
    X : numpy.ndarray (n_samples, n_var)
        Original population in decision space (including violations)

    Xb : numpy.ndarray (n_samples, n_var)
        Feasible reference points in decision space

    xl : numpy.ndarray (n_var,)
        Lower bounds for decision variables

    xu : numpy.ndarray (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.ndarray (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + (Xb[i, j] - XL[i, j]) / 2

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - (XU[i, j] - Xb[i, j]) / 2

    return X


def to_bounds(X, Xb, xl, xu):
    """Repair strategy that forces violated elements to problem bounds

    Parameters
    ----------
    X : numpy.ndarray (n_samples, n_var)
        Original population in decision space (including violations)

    Xb : numpy.ndarray (n_samples, n_var)
        Feasible reference points in decision space

    xl : numpy.ndarray (n_var,)
        Lower bounds for decision variables

    xu : numpy.ndarray (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.ndarray (n_samples, n_var)
        Repaired population
    """

    xl = np.array(xl)
    xu = np.array(xu)

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j]

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j]

    return X


def rand_init(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements to problem domain

    Parameters
    ----------
    X : numpy.ndarray (n_samples, n_var)
        Original population in decision space (including violations)

    Xb : numpy.ndarray (n_samples, n_var)
        Feasible reference points in decision space

    xl : numpy.ndarray (n_var,)
        Lower bounds for decision variables

    xu : numpy.ndarray (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.ndarray (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.random.random(len(i)) * (XU[i, j] - XL[i, j])

    return X


def squared_bounce_back(X, Xb, xl, xu):
    """Repair strategy that radomly re-initializes violated elements between reference points and bounds
    closer to bounds

    Parameters
    ----------
    X : numpy.ndarray (n_samples, n_var)
        Original population in decision space (including violations)

    Xb : numpy.ndarray (n_samples, n_var)
        Feasible reference points in decision space

    xl : numpy.ndarray (n_var,)
        Lower bounds for decision variables

    xu : numpy.ndarray (n_var,)
        Upper bounds for decision variables

    Returns
    -------
    numpy.ndarray (n_samples, n_var)
        Repaired population
    """

    XL = xl[None, :].repeat(len(X), axis=0)
    XU = xu[None, :].repeat(len(X), axis=0)

    i, j = np.where(X < XL)
    if len(i) > 0:
        X[i, j] = XL[i, j] + np.square(np.random.random(len(i))) * (Xb[i, j] - XL[i, j])

    i, j = np.where(X > XU)
    if len(i) > 0:
        X[i, j] = XU[i, j] - np.square(np.random.random(len(i))) * (XU[i, j] - Xb[i, j])

    return X


def normalize_fun(fun):

    fmin = fun.min(axis=0)
    fmax = fun.max(axis=0)
    den = fmax - fmin

    den[den <= 1e-16] = 1.0

    return (fun - fmin)/den


REPAIRS = {"bounce-back": bounce_back,
           "midway": midway,
           "rand-init": rand_init,
           "to-bounds": to_bounds}


# Native
from typing import Optional

# pymoo imports
from pymoo.core.algorithm import Algorithm
from pymoo.core.duplicate import DefaultDuplicateElimination, NoDuplicateElimination
from pymoo.core.initialization import Initialization
from pymoo.core.infill import InfillCriterion
from pymoo.core.population import Population
from pymoo.core.repair import NoRepair, Repair
from pymoo.core.survival import Survival
from pymoo.core.sampling import Sampling
from pymoo.core.mating import Mating


# =========================================================================================================
# Implementation
# =========================================================================================================

class EvolutionaryAlgorithm(Algorithm):

    def __init__(
        self,
        pop_size: Optional[int] = None,
        sampling: Optional[Sampling] = None,
        mating: Optional[Mating] = None,
        survival: Optional[Survival] = None,
        n_offsprings: Optional[int] = None,
        eliminate_duplicates: bool = True,
        repair: Optional[Repair] = None,
        advance_after_initial_infill: bool = False,
        **kwargs
    ):
        """Base class for Evolutionary Algorithms

        Parameters
        ----------
        pop_size : int, optional
            Population size, by default None

        sampling : Sampling, optional
            pymoo Sampling instance, by default None

        mating : InfillCriterion, optional
            pymoo mating operator, by default None

        survival : Survival, optional
            pymoo survival operator, by default None

        n_offsprings : int, optional
            Number of offspring individuals created at each generation, by default None

        eliminate_duplicates : DuplicateElimination | bool | None, optional
            Eliminate duplicates in mating, by default True

        repair : Repair, optional
            pymoo repair operator. In the algorithm level it should be called when sampling.
            It is recommended that mating operators also have some repair associated with. By default None

        advance_after_initial_infill : bool, optional
            Either or not apply survival after initialization, by default False
        """

        super().__init__(**kwargs)

        # the population size used
        self.pop_size = pop_size

        # whether the algorithm should be advanced after initialization of not
        self.advance_after_initial_infill = advance_after_initial_infill

        # the survival for the genetic algorithm
        self.survival = survival

        # number of offsprings to generate through recombination
        self.n_offsprings = n_offsprings

        # if the number of offspring is not set - equal to population size
        if self.n_offsprings is None:
            self.n_offsprings = pop_size

        # set the duplicate detection class - a boolean value chooses the default duplicate detection
        if isinstance(eliminate_duplicates, bool):
            if eliminate_duplicates:
                self.eliminate_duplicates = DefaultDuplicateElimination()
            else:
                self.eliminate_duplicates = NoDuplicateElimination()
        else:
            self.eliminate_duplicates = eliminate_duplicates

        # simply set the no repair object if it is None
        self.repair = repair if repair is not None else NoRepair()

        self.initialization = Initialization(
            sampling, repair=self.repair,
            eliminate_duplicates=self.eliminate_duplicates,
        )

        if mating is None:
            mating = InfillCriterion(
                repair=self.repair,
                eliminate_duplicates=self.eliminate_duplicates,
        )
        self.mating = mating

        # other run specific data updated whenever solve is called - to share them in all algorithms
        self.n_gen = None
        self.pop = None
        self.off = None

    def _initialize_infill(self):
        pop = self.initialization.do(self.problem, self.pop_size, algorithm=self)
        return pop

    def _initialize_advance(self, infills=None, **kwargs):
        if self.advance_after_initial_infill:
            self.pop = self.survival.do(self.problem, infills, n_survive=len(infills), algorithm=self, **kwargs)

    def _infill(self):

        # do the mating using the current population
        off = self.mating.do(self.problem, self.pop, self.n_offsprings, algorithm=self)

        # if the mating could not generate any new offspring (duplicate elimination might make that happen)
        if len(off) == 0:
            self.termination.force_termination = True
            return

        # if not the desired number of offspring could be created
        elif len(off) < self.n_offsprings:
            if self.verbose:
                print("WARNING: Mating could not produce the required number of (unique) offsprings!")

        return off

    def _advance(self, infills=None, **kwargs):

        # the current population
        pop = self.pop

        # merge the offsprings with the current population
        if infills is not None:
            pop = Population.merge(self.pop, infills)

        # execute the survival to find the fittest solutions
        self.pop = self.survival.do(self.problem, pop, n_survive=self.pop_size, algorithm=self, **kwargs)



import pytest

from pymoo.optimize import minimize
from pymoo.problems import get_problem
from pymoo.indicators.igd import IGD
from pymoode.algorithms import GDE3, NSDE
from pymoode.operators.dex import DEX
from pymoode.operators.dem import DEM
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.operators.mutation.pm import PM
from pymoode.survival import RankAndCrowding, ConstrRankAndCrowding


@pytest.mark.parametrize('survival', [RankAndCrowding, ConstrRankAndCrowding])
@pytest.mark.parametrize('crowding_func', ["mnn", "2nn", "cd", "pcd", "ce"])
def test_multi_run(survival, crowding_func):
    
    problem = get_problem("truss2d")

    NGEN = 50
    POPSIZE = 50
    SEED = 5
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=survival(crowding_func=crowding_func))

    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    assert len(res_gde3.opt) > 0


@pytest.mark.parametrize('crossover', [DEX(), DEM()])
def test_multi_frankstein(crossover):
    
    problem = get_problem("truss2d")

    NGEN = 100
    POPSIZE = 100
    SEED = 5
    
    frank = NSGA2(pop_size=POPSIZE, crossover=crossover)

    res_frank = minimize(problem,
                        frank,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    assert len(res_frank.opt) > 0


def test_gde3_pm_run():
    
    problem = get_problem("truss2d")

    NGEN = 50
    POPSIZE = 50
    SEED = 5
    
    gde3pm = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="pcd"), genetic_mutation=PM())

    res_gde3pm = minimize(problem,
                          gde3pm,
                          ('n_gen', NGEN),
                          seed=SEED,
                          save_history=False,
                          verbose=False)
    
    assert len(res_gde3pm.opt) > 0
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="pcd"))
    
    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    assert len(sum(res_gde3pm.F - res_gde3.F)) >= 1e-3
    

def test_multi_perf():
    
    problem = get_problem("truss2d")
    igd = IGD(pf=problem.pareto_front(), zero_to_one=True)
    
    NGEN = 250
    POPSIZE = 100
    SEED = 5
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="cd"))

    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    igd_gde3 = igd.do(res_gde3.F)
    assert abs(igd_gde3 - 0.005859828655308572) <= 1e-8
    
    gde3p = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="pcd"))

    res_gde3p = minimize(problem,
                        gde3p,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    igd_gde3p = igd.do(res_gde3p.F)
    assert abs(igd_gde3p - 0.004744463013355145) <= 1e-8
    
    nsde = NSDE(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="pcd"))
        
    res_nsde = minimize(problem,
                        nsde,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    igd_nsde = igd.do(res_nsde.F)
    assert abs(igd_nsde - 0.004562068055351625) <= 1e-8


# Native
from typing import Optional

# External
import numpy as np

# pymoo imports
from pymoo.core.survival import Survival
from pymoo.core.population import Population
from pymoo.core.problem import Problem


# =========================================================================================================
# Implementation
# =========================================================================================================

class BaseFitnessSurvival(Survival):

    def __init__(self) -> None:
        super().__init__(filter_infeasible=False)

    def _do(
        self,
        problem: Problem,
        pop: Population,
        n_survive: Optional[int]=None,
        **kwargs
    ):
        return pop[:n_survive]


class FitnessSurvival(Survival):

    def __init__(self) -> None:
        super().__init__(filter_infeasible=False)

    def _do(
        self,
        problem: Problem,
        pop: Population,
        n_survive: Optional[int]=None,
        **kwargs
    ):
        F, cv = pop.get("F", "cv")
        assert F.shape[1] == 1, "FitnessSurvival can only used for single objective single!"
        S = np.lexsort([F[:, 0], cv])
        pop.set("rank", np.argsort(S))
        return pop[S[:n_survive]]



import pytest

import numpy as np
from pymoo.optimize import minimize
from pymoo.problems import get_problem
from pymoode.performance import SpacingIndicator
from pymoode.algorithms import GDE3
from pymoode.survival import RankAndCrowding


def test_spacing_values():
    
    F = np.array([
        [1.2, 7.8],
        [2.8, 5.1],
        [4.0, 2.8],
        [7.0, 2.2],
        [8.4, 1.2]
    ])
    
    assert abs(SpacingIndicator().do(F) - 0.73) <= 1e-3
    
    F = np.array([
        [1.0, 7.5],
        [1.1, 5.5],
        [2.0, 5.0],
        [3.0, 4.0],
        [4.0, 2.8],
        [5.5, 2.5],
        [6.8, 2.0],
        [8.4, 1.2]
    ])

    assert abs(SpacingIndicator().do(F) - 0.316) <= 1e-3


def test_spacing_zero_to_one():
    
    problem = get_problem("truss2d")
    sp_from_pf = SpacingIndicator(pf=problem.pareto_front(), zero_to_one=True)
    sp_from_ref_points = SpacingIndicator(zero_to_one=True, nadir=problem.nadir_point(), ideal=problem.ideal_point())
    sp_no_scale = SpacingIndicator(zero_to_one=False)

    NGEN = 250
    POPSIZE = 100
    SEED = 5
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 0.9), de_repair="bounce-back",
                survival=RankAndCrowding(crowding_func="pcd"))

    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    assert sp_from_pf.do(res_gde3.F) > 0
    assert sp_from_ref_points.do(res_gde3.F) > 0
    assert sp_no_scale.do(res_gde3.F) > 0
    assert abs(sp_from_pf.do(res_gde3.F) - sp_from_ref_points.do(res_gde3.F)) <= 1e-6


# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

# This was implemented using the full distances matrix
# Other strategies can be more efficient depending on the population size and number of objectives
# This approach was the most promising for N = 3
# I believe for a large number of objectives M, some strategy based on upper bounds for distances would be helpful
# Those interested in contributing please contact me at bruscalia12@gmail.com


import numpy as np

from pymoode.cython.utils cimport c_get_drop, c_get_argmin, c_get_argmax, c_normalize_array

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.set cimport set as cpp_set


cdef extern from "math.h":
    double HUGE_VAL


def calc_mnn(double[:, :] X, int n_remove=0):

    cdef:
        int N, M, n
        cpp_set[int] extremes
        vector[int] extremes_min, extremes_max

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
    else:
        n_remove = N - M

    extremes_min = c_get_argmin(X)
    extremes_max = c_get_argmax(X)

    extremes = cpp_set[int]()

    for n in extremes_min:
        extremes.insert(n)

    for n in extremes_max:
        extremes.insert(n)

    X = c_normalize_array(X, extremes_max, extremes_min)

    return c_calc_mnn(X, n_remove, N, M, extremes)


def calc_2nn(double[:, :] X, int n_remove=0):

    cdef:
        int N, M, n
        cpp_set[int] extremes
        vector[int] extremes_min, extremes_max

    N = X.shape[0]
    M = X.shape[1]

    if n_remove <= (N - M):
        if n_remove < 0:
            n_remove = 0
    else:
        n_remove = N - M

    extremes_min = c_get_argmin(X)
    extremes_max = c_get_argmax(X)

    extremes = cpp_set[int]()

    for n in extremes_min:
        extremes.insert(n)

    for n in extremes_max:
        extremes.insert(n)

    X = c_normalize_array(X, extremes_max, extremes_min)

    M = 2

    return c_calc_mnn(X, n_remove, N, M, extremes)


cdef double[:] c_calc_mnn(double[:, :] X, int n_remove, int N, int M, cpp_set[int] extremes):

    cdef:
        int n, mm, i, j, n_removed, k, MM
        double dij
        cpp_set[int] calc_items
        cpp_set[int] H
        double[:, :] D
        double[:] d
        int[:, :] Mnn

    # Return infinity if not enough points
    if N <= M:
        return np.full(N, HUGE_VAL, dtype=np.double)

    # Define items to calculate distances
    calc_items = cpp_set[int]()
    for n in range(N):
        calc_items.insert(n)
    for n in extremes:
        calc_items.erase(n)

    # Define remaining items to evaluate
    H = cpp_set[int]()
    for n in range(N):
        H.insert(n)

    # Instantiate distances array
    _D = np.empty((N, N), dtype=np.double)
    D = _D[:, :]

    # Shape of X
    MM = X.shape[1]

    # Fill values on D
    for i in range(N - 1):
        D[i, i] = 0.0

        for j in range(i + 1, N):

            dij = 0
            for mm in range(MM):
                dij = dij + (X[j, mm] - X[i, mm]) * (X[j, mm] - X[i, mm])

            D[i, j] = dij
            D[j, i] = D[i, j]

    D[N-1, N-1] = 0.0

    # Initialize
    n_removed = 0

    # Initialize neighbors and distances
    # _Mnn = np.full((N, M), -1, dtype=np.intc)
    _Mnn = np.argpartition(D, range(1, M+1), axis=1)[:, 1:M+1].astype(np.intc)
    dd = np.full((N,), HUGE_VAL, dtype=np.double)

    Mnn = _Mnn[:, :]
    d = dd[:]

    # Obtain distance metrics
    c_calc_d(d, Mnn, D, calc_items, M)

    # While n_remove not acheived (no need to recalculate if only one item should be removed)
    while n_removed < (n_remove - 1):

        # Obtain element to drop
        k = c_get_drop(d, H)
        H.erase(k)

        # Update index
        n_removed = n_removed + 1

        # Get items to be recalculated
        calc_items = c_get_calc_items(Mnn, H, k, M)
        for n in extremes:
            calc_items.erase(n)

        # Fill in neighbors and distance matrix
        c_calc_mnn_iter(
                X,
                Mnn,
                D,
                N, M,
                calc_items,
                H
            )

        # Obtain distance metrics
        c_calc_d(d, Mnn, D, calc_items, M)

    return dd


cdef c_calc_mnn_iter(
    double[:, :] X,
    int[:, :] Mnn,
    double[:, :] D,
    int N, int M,
    cpp_set[int] calc_items,
    cpp_set[int] H
    ):

    cdef:
        int i, j, m

    # Iterate over items to calculate
    for i in calc_items:

        # Iterate over elements in X
        for j in H:

            # Go to next if same element
            if (j == i):
                continue

            # Replace at least the last neighbor
            elif (D[i, j] <= D[i, Mnn[i, M-1]]) or (Mnn[i, M-1] == -1):

                # Iterate over current values
                for m in range(M):

                    # Set to current if unassigned
                    if (Mnn[i, m] == -1):

                        # Set last neighbor to index
                        Mnn[i, m] = j
                        break

                    # Break if checking already corresponding index
                    elif (j == Mnn[i, m]):
                        break

                    # Distance satisfies condition
                    elif (D[i, j] <= D[i, Mnn[i, m]]):

                        # Replace higher values
                        Mnn[i, m + 1:] = Mnn[i, m:-1]

                        # Replace current value
                        Mnn[i, m] = j
                        break


# Calculate crowding metric
cdef c_calc_d(double[:] d, int[:, :] Mnn, double[:, :] D, cpp_set[int] calc_items, int M):

    cdef:
        int i, m

    for i in calc_items:

        d[i] = 1
        for m in range(M):
            d[i] = d[i] * D[i, Mnn[i, m]]


# Returns indexes of items to be recalculated after removal
cdef cpp_set[int] c_get_calc_items(
    int[:, :] Mnn,
    cpp_set[int] H,
    int k, int M):

    cdef:
        int i, m
        cpp_set[int] calc_items

    calc_items = cpp_set[int]()

    for i in H:

        for m in range(M):

            if Mnn[i, m] == k:

                Mnn[i, m:-1] = Mnn[i, m + 1:]
                Mnn[i, M-1] = -1

                calc_items.insert(i)

    return calc_items



# Native
import warnings

# External
import numpy as np

# pymoo imports
from pymoo.core.selection import Selection
from pymoo.core.population import Population
from pymoo.core.problem import Problem


# =========================================================================================================
# Implementation
# =========================================================================================================


# This is the core differential evolution selection class
class DES(Selection):

    def __init__(self, variant: str, **kwargs):
        """Differential Evolution parent selection class

        Parameters
        ----------
        variant : str, optional
            Differential evolution strategy. Must be a string in the format: "DE/selection/n/crossover", in which, n in an integer of number of difference vectors, and crossover is either 'bin' or 'exp'. Selection variants are:

                - 'ranked'
                - 'rand'
                - 'best'
                - 'current-to-best'
                - 'current-to-rand'
                - 'rand-to-best'
        """

        super().__init__()
        self.variant = variant

    def _do(
        self,
        problem: Problem,
        pop: Population,
        n_select: int,
        n_parents: int,
        **kwargs
    ):

        # Obtain number of elements in population
        n_pop = len(pop)
        if n_pop != n_select:
            _warn_n_select()

        if self.variant == "ranked":
            """Proposed by Zhang et al. (2021). doi.org/10.1016/j.asoc.2021.107317"""
            P = self._ranked(pop, n_select, n_parents)

        elif self.variant == "best":
            P = self._best(pop, n_select, n_parents)

        elif self.variant == "current-to-best":
            P = self._current_to_best(pop, n_select, n_parents)

        elif self.variant == "current-to-rand":
            P = self._current_to_rand(pop, n_select, n_parents)

        elif self.variant == "rand-to-best":
            P = self._rand_to_best(pop, n_select, n_parents)

        elif self.variant == "rand":
            P = self._rand(pop, n_select, n_parents)

        else:
            _warn_variant()
            P = self._rand(pop, n_select, n_parents)

        return P

    def _rand(self, pop, n_select, n_parents, **kwargs):

        # len of pop
        n_pop = len(pop)

        # Base form
        P = np.empty([n_select, n_parents], dtype=int)

        # Fill target vector with corresponding parent
        target = np.arange(n_pop)[:n_select]

        # Fill next columns in loop
        for j in range(n_parents):

            P[:, j] = np.random.choice(n_pop, n_select)
            reselect = get_reselect(P, target, j)

            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = get_reselect(P, target, j)

        return P

    def _best(self, pop, n_select, n_parents, **kwargs):

        # len of pop
        n_pop = len(pop)

        # Base form
        P = np.empty([n_select, n_parents], dtype=int)

        # Fill target vector with corresponding parent
        target = np.arange(n_pop)[:n_select]

        # Fill first column with best candidate
        P[:, 0] = 0

        # Fill next columns in loop
        for j in range(1, n_parents):

            P[:, j] = np.random.choice(n_pop, n_select)
            reselect = get_reselect(P, target, j)

            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = get_reselect(P, target, j)

        return P

    def _current_to_best(self, pop, n_select, n_parents, **kwargs):

        # len of pop
        n_pop = len(pop)

        # Base form
        P = np.empty([n_select, n_parents], dtype=int)

        # Fill target vector with corresponding parent
        target = np.arange(n_pop)[:n_select]

        # Fill first column with current candidate
        P[:, 0] = np.arange(n_pop)

        # Fill first direction from current
        P[:, 2] = np.arange(n_pop)

        # Towards best
        P[:, 1] = 0

        # Fill next columns in loop
        for j in range(3, n_parents):

            P[:, j] = np.random.choice(n_pop, n_select)
            reselect = get_reselect(P, target, j)

            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = get_reselect(P, target, j)

        return P

    def _current_to_rand(self, pop, n_select, n_parents, **kwargs):

        # len of pop
        n_pop = len(pop)

        # Base form
        P = np.empty([n_select, n_parents], dtype=int)

        # Fill target vector with corresponding parent
        target = np.arange(n_pop)[:n_select]

        # Fill first column with current candidate
        P[:, 0] = np.arange(n_pop)

        # Fill first direction from current
        P[:, 2] = np.arange(n_pop)

        # Towards random
        P[:, 1] = np.random.choice(n_pop, n_select)
        reselect = get_reselect(P, target, 1)

        while np.any(reselect):
            P[reselect, 1] = np.random.choice(n_pop, reselect.sum())
            reselect = get_reselect(P, target, 1)

        # Fill next columns in loop
        for j in range(3, n_parents):

            P[:, j] = np.random.choice(n_pop, n_select)
            reselect = get_reselect(P, target, j)

            while np.any(reselect):
                P[reselect, j] = np.random.choice(n_pop, reselect.sum())
                reselect = get_reselect(P, target, j)

        return P

    def _rand_to_best(self, pop, n_select, n_parents, **kwargs):

        PB = self._best(pop, n_select, n_parents, **kwargs)
        P = PB.copy()

        P[:, 0] = PB[:, 1]
        P[:, 1] = PB[:, 0]

        return P

    def _ranked(self, pop, n_select, n_parents, **kwargs):

        P = self._rand(pop, n_select, n_parents, **kwargs)
        P[:, 1:] = rank_sort(P[:, 1:], pop)

        return P


def _warn_n_select():
    warnings.warn(
        "DE parent selection is supposed to work with n_select as the population size",
        UserWarning
    )


def _warn_variant():
    warnings.warn(
        "Unknown selection variant; using 'rand' instead",
        UserWarning
    )


def get_reselect(P: np.ndarray, target: np.ndarray, j: int):
    return (P[:, j] == target) | (P[:, j].reshape([-1, 1]) == P[:, :j]).any(axis=1)


def ranks_from_cv(pop: Population):

    ranks = pop.get("rank")
    cv_elements = ranks == None

    if np.any(cv_elements):
        ranks[cv_elements] = np.arange(len(pop))[cv_elements]

    return ranks


def rank_sort(P: np.ndarray, pop: Population):

    ranks = ranks_from_cv(pop)

    sorted = np.argsort(ranks[P], axis=1, kind="stable")
    S = np.take_along_axis(P, sorted, axis=1)

    P[:, 0] = S[:, 0]

    n_diffs = int((P.shape[1] - 1) / 2)

    for j in range(1, n_diffs + 1):
        P[:, 2*j - 1] = S[:, j]
        P[:, 2*j] = S[:, -j]

    return P


def reiforce_directions(P: np.ndarray, pop: Population):

    ranks = ranks_from_cv(pop)

    ranks = ranks[P]
    S = P.copy()

    n_diffs = int(P.shape[1] / 2)

    for j in range(0, n_diffs):
        bad_directions = ranks[:, 2*j] > ranks[:, 2*j + 1]
        P[bad_directions, 2*j] = S[bad_directions, 2*j + 1]
        P[bad_directions, 2*j + 1] = S[bad_directions, 2*j]

    return P



import pytest

from pymoo.optimize import minimize
from pymoo.problems import get_problem
from pymoo.indicators.igd import IGD
from pymoode.algorithms import NSDER
from pymoode.algorithms import GDE3
from pymoode.survival import RankAndCrowding
from pymoo.util.ref_dirs import get_reference_directions

import numpy as np

@pytest.mark.parametrize('selection', ["rand", "current-to-rand", "ranked"])
@pytest.mark.parametrize('crossover', ["bin", "exp"])
@pytest.mark.parametrize('crowding_func', ["mnn", "2nn"])
def test_many_run(selection, crossover, crowding_func):
    
    problem = get_problem("dtlz2")
    
    NGEN = 150
    POPSIZE = 136
    SEED = 5
    
    gde3 = GDE3(pop_size=POPSIZE, variant=f"DE/{selection}/1/{crossover}", CR=0.2, F=(0.0, 1.0), gamma=1e-4,
                survival=RankAndCrowding(crowding_func=crowding_func))

    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    assert len(res_gde3.opt) > 0
    

def test_many_perf():
    
    np.random.seed(3)
    assert abs(np.random.rand() - 0.5507979025745755) <= 1e-8
    
    problem = get_problem("dtlz2")
    ref_dirs = get_reference_directions("das-dennis", 3, n_partitions=15)
    igd = IGD(pf=problem.pareto_front(), zero_to_one=True)
    
    NGEN = 150
    POPSIZE = 136
    SEED = 5
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.2, F=(0.0, 1.0), gamma=1e-4,
                survival=RankAndCrowding(crowding_func="mnn"))

    res_gde3 = minimize(problem,
                        gde3,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    igd_gde3 = igd.do(res_gde3.F)
    assert abs(igd_gde3 - 0.04011488503871424) <= 1e-8
    
    nsder = NSDER(ref_dirs=ref_dirs, pop_size=POPSIZE, variant="DE/rand/1/bin", CR=0.5, F=(0.0, 1.0), gamma=1e-4)
    
    res_nsder = minimize(problem,
                        nsder,
                        ('n_gen', NGEN),
                        seed=SEED,
                        save_history=False,
                        verbose=False)
    
    igd_nsder = igd.do(res_nsder.F)
    assert abs(igd_nsder - 0.004877000918527632) <= 1e-8
    
    
    
    


# Native
from abc import abstractmethod
from typing import Optional, Union

# External
import numpy as np

# pymoo imports
from pymoo.core.population import Population
from pymoo.core.problem import Problem

# pymoode imports
from pymoode.operators.deop import DifferentialOperator


# =========================================================================================================
# Implementation
# =========================================================================================================

class DifferentialCrossover(DifferentialOperator):

    def __init__(self,
                 variant="bin",
                 CR=0.7,
                 at_least_once=True,
                 **kwargs):
        """Differential evolution crossover
        (DE mutation is considered a part of this operator)

        Parameters
        ----------
        variant : str | callable, optional
            Crossover variant. Must be either "bin", "exp", or callable. By default "bin".
            If callable, it has the form:
            ``cross_function(n_matings, n_var, CR, at_least_once=True)``

        CR : float, optional
            Crossover parameter. Defined in the range [0, 1]
            To reinforce mutation, use higher values. To control convergence speed, use lower values.

        at_least_once : bool, optional
            Either or not offsprings must inherit at least one attribute from mutant vectors, by default True
        """

        # __init__ operator
        super().__init__(n_parents=2, **kwargs)

        self.CR = CR
        self.variant = variant
        self.at_least_once = at_least_once

    def do(
        self,
        problem: Problem,
        pop: Population,
        parents: Optional[Union[Population, np.ndarray]] = None,
        **kwargs
    ):

        # Convert pop if parents is not None
        pop, X = self.default_prepare(pop, parents)

        # Create child vectors
        U = self._do(problem, X, **kwargs)

        return Population.new("X", U)

    @abstractmethod
    def _do(self, problem: Problem, X: np.ndarray, **kwargs):
        pass


class DEX(DifferentialCrossover):

    def __init__(self,
                 variant="bin",
                 CR=0.7,
                 at_least_once=True,
                 **kwargs):

        super().__init__(
            variant=variant, CR=CR,
            at_least_once=at_least_once,
            **kwargs,
        )

        if self.variant == "bin":
            self.cross_function = cross_binomial
        elif self.variant == "exp":
            self.cross_function = cross_exp
        elif hasattr(self.variant, "__call__"):
            self.cross_function = self.variant
        else:
            raise ValueError("Crossover variant must be either 'bin', 'exp', or callable")

    def _do(self, problem: Problem, X: np.ndarray, **kwargs):

        # Decompose input vector
        V = X[1]
        X_ = X[0]
        U = np.array(X_, copy=True)

        # About X
        n_matings, n_var = X_.shape

        # Mask
        M = self.cross_function(n_matings, n_var, self.CR, self.at_least_once)
        U[M] = V[M]

        return U


# =========================================================================================================
# Crossovers
# =========================================================================================================


# From version 0.5.0 of pymoo
def row_at_least_once_true(M: np.ndarray):

    _, d = M.shape

    for k in np.where(~np.any(M, axis=1))[0]:
        M[k, np.random.randint(d)] = True

    return M


def cross_binomial(n_matings, n_var, prob, at_least_once=True):

    M = np.random.random((n_matings, n_var)) < prob

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


def cross_exp(n_matings, n_var, prob, at_least_once=True):

    # the mask do to the crossover
    M = np.full((n_matings, n_var), False)

    # start point of crossover
    s = np.random.randint(0, n_var, size=n_matings)

    # create for each individual the crossover range
    for i in range(n_matings):

        # the actual index where we start
        start = s[i]
        for j in range(n_var):

            # the current position where we are pointing to
            current = (start + j) % n_var

            # replace only if random value keeps being smaller than CR
            if np.random.random() <= prob:
                M[i, current] = True
            else:
                break

    if at_least_once:
        M = row_at_least_once_true(M)

    return M



# distutils: language = c++
# cython: language_level=2, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.set cimport set as cpp_set


cdef extern from "math.h":
    double HUGE_VAL


def calc_spacing_distances(double[:, :] X):

    return c_calc_spacing_distances(X)


cdef c_calc_spacing_distances(double[:, :] X):

    cdef:
        int N, M, n, m
        cpp_set[int] calc_items
        double[:, :] D
        double[:] d
        double dij, _dijm

    N = X.shape[0]
    M = X.shape[1]

    # Initialize neighbors and distances
    _D = np.full((N, N), -1.0, dtype=np.double)
    dd = np.full((N,), HUGE_VAL, dtype=np.double)

    D = _D[:, :]
    d = dd[:]

    # Iterate over items to calculate
    for i in range(N):

        # Iterate over elements in X
        for j in range(N):

            # Calculate distance if elements satisfy rule
            if ((j != i) and (D[i, j] <= d[i])):

                # Calculate if different from -1
                if D[i, j] == -1.0:

                    # Squared distance
                    dij = 0
                    for m in range(M):
                        _dijm = X[j, m] - X[i, m]

                        if _dijm >= 0:
                            dij = dij + _dijm

                        else:
                            dij = dij - _dijm

                    # Fill values
                    D[i, j] = dij
                    D[j, i] = D[i, j]

                else:
                    dij = D[i, j]

                # Check is any should be replaced
                if (dij <= d[i]):
                    d[i] = dij

    return dd



import pytest

from pymoo.optimize import minimize
from pymoo.problems import get_problem
from pymoo.operators.mutation.pm import PM
from pymoode.algorithms import GDE3
from pymoode.algorithms import DE


@pytest.mark.parametrize('selection', ["rand", "best", "current-to-best", "current-to-rand", "ranked"])
@pytest.mark.parametrize('crossover', ["bin", "exp"])
@pytest.mark.parametrize('de_repair', ["bounce-back", "midway", "rand-init", "to-bounds"])
def test_de_run(selection, crossover, de_repair):
    problem = get_problem("rastrigin")
    
    NGEN = 50
    POPSIZE = 20
    SEED = 3
    
    #DE Parameters
    CR = 0.5
    F = (0.3, 1.0)

    de = DE(pop_size=POPSIZE, variant=f"DE/{selection}/1/{crossover}", CR=CR, F=F, de_repair=de_repair)

    res_de = minimize(problem,
                    de,
                    ('n_gen', NGEN),
                    seed=SEED,
                    save_history=False,
                    verbose=False)

    assert len(res_de.opt) > 0


def test_de_pm_run():
    
    problem = get_problem("rastrigin")
    
    NGEN = 50
    POPSIZE = 20
    SEED = 3
    
    #DE Parameters
    CR = 0.5
    F = (0.3, 1.0)

    de = DE(pop_size=POPSIZE, variant=f"DE/rand/1/bin", CR=CR, F=F)

    res_de = minimize(problem,
                      de,
                      ('n_gen', NGEN),
                      seed=SEED,
                      save_history=False,
                      verbose=False)
    
    depm = DE(pop_size=POPSIZE, variant=f"DE/rand/1/bin", CR=CR, F=F, genetic_mutation=PM())

    res_pm = minimize(problem,
                      depm,
                      ('n_gen', NGEN),
                      seed=SEED,
                      save_history=False,
                      verbose=False)

    assert len(res_pm.opt) > 0
    assert len(sum(res_de.pop.get("F") - res_pm.pop.get("F"))) >= 1e-6
    

def test_de_perf():
    problem = get_problem("rastrigin")
    
    NGEN = 100
    POPSIZE = 20
    SEED = 3
    
    #DE Parameters
    CR = 0.5
    F = (0.3, 1.0)

    de = DE(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=CR, F=F)

    res_de = minimize(problem,
                    de,
                    ('n_gen', NGEN),
                    seed=SEED,
                    save_history=False,
                    verbose=False)

    assert len(res_de.opt) > 0
    assert res_de.F <= 1e-6
    
    gde3 = GDE3(pop_size=POPSIZE, variant="DE/rand/1/bin", CR=CR, F=F)
    
    res_gde3 = minimize(problem,
                    gde3,
                    ('n_gen', NGEN),
                    seed=SEED,
                    save_history=False,
                    verbose=False)
    
    assert res_gde3.F <= 1e-6

